/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2010-2014 QuakeSpasm developers

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef spingle__common_h
#define spingle__common_h

#include "sdl/endian_sdl.h"

#define q_minfunc(typ, name) \
	static inline typ name(typ a, typ b) {return a < b ? a : b;}

#define q_maxfunc(typ, name) \
	static inline typ name(typ a, typ b) {return a > b ? a : b;}

#define CLAMPfunc(typ, name) \
	static inline typ name(typ min, typ x, typ max) \
	{ \
		return x < min ? min : x > max ? max : x; \
	}

#define GenericTypes(func, name) \
	func(char,     name##chr) \
	func(byte,     name##byt) \
	func(int16_t,  name##i16) \
	func(int32_t,  name##i32) \
	func(int64_t,  name##i64) \
	func(uint16_t, name##u16) \
	func(uint32_t, name##u32) \
	func(uint64_t, name##u64) \
	func(float,    name##f) \
	func(double,   name##d)

#define GenericDecl(typ, name) , typ: name

GenericTypes(q_minfunc, q_min)
GenericTypes(q_maxfunc, q_max)
GenericTypes(CLAMPfunc, CLAMP)

#define CLAMP(minval, x, maxval) \
	_Generic((x) GenericTypes(GenericDecl, CLAMP))(minval, x, maxval)

#define q_min(a, b) _Generic((a) GenericTypes(GenericDecl, q_min))(a, b)
#define q_max(a, b) _Generic((a) GenericTypes(GenericDecl, q_max))(a, b)

#undef q_minfunc
#undef q_maxfunc
#undef CLAMPfunc

typedef struct sizebuf_s
{
	bool    allowoverflow; // if false, do a Sys_Error
	bool    overflowed;    // set to true if the buffer size failed
	byte    *data;
	int32_t maxsize;
	int32_t cursize;
} sizebuf_t;

void SZ_Alloc(sizebuf_t *buf, int32_t startsize);
void SZ_Free(sizebuf_t *buf);
void SZ_Clear(sizebuf_t *buf);
void *SZ_GetSpace(sizebuf_t *buf, int32_t length);
void SZ_Write(sizebuf_t *buf, const void *data, int32_t length);
void SZ_Print(sizebuf_t *buf, const char *data);        // strcats onto the sizebuf

//============================================================================

typedef struct link_s
{
	struct link_s   *prev, *next;
} link_t;


void ClearLink(link_t *l);
void RemoveLink(link_t *l);
void InsertLinkBefore(link_t *l, link_t *before);
void InsertLinkAfter(link_t *l, link_t *after);

//============================================================================

#define BytesX(typ, name) \
	static inline typ Bytes##name(byte const *bytes) \
	{ \
		typ v; \
		memcpy(&v, bytes, sizeof(typ)); \
		return v; \
	}

#define ReadXX(endian, typ, name) \
	static inline typ Read##endian##name(byte const **bytes) \
	{ \
		typ v = endian##name(Bytes##name(*bytes)); \
		*bytes += sizeof(typ); \
		return v; \
	}

#define Reader(typ, name) \
	BytesX(typ, name) \
	ReadXX(Big, typ, name) \
	ReadXX(Little, typ, name)

Reader(int16_t, Short)
Reader(int32_t, Long)
Reader(float,   Float)

static inline void ReadSkip(byte const **bytes, size_t n)
{
	*bytes += n;
}

static inline void ReadCopy(void *out, byte const **bytes, size_t n)
{
	memcpy(out, *bytes, n);
	*bytes += n;
}

#undef BytesX
#undef ReadXX
#undef Reader

//============================================================================

void MSG_WriteChar(sizebuf_t *sb, int32_t c);
void MSG_WriteByte(sizebuf_t *sb, int32_t c);
void MSG_WriteShort(sizebuf_t *sb, int32_t c);
void MSG_WriteLong(sizebuf_t *sb, int32_t c);
void MSG_WriteFloat(sizebuf_t *sb, float f);
void MSG_WriteString(sizebuf_t *sb, const char *s);
void MSG_WriteCoord(sizebuf_t *sb, float f, uint32_t flags);
void MSG_WriteAngle(sizebuf_t *sb, float f, uint32_t flags);
void MSG_WriteAngle16(sizebuf_t *sb, float f, uint32_t flags);  //johnfitz

extern  int32_t                 msg_readcount;
extern  bool    msg_badread;            // set if a read goes beyond end of message

void MSG_BeginReading(void);
int32_t MSG_ReadChar(void);
int32_t MSG_ReadByte(void);
int32_t MSG_ReadShort(void);
int32_t MSG_ReadLong(void);
float MSG_ReadFloat(void);
const char *MSG_ReadString(void);

float MSG_ReadCoord(uint32_t flags);
float MSG_ReadAngle(uint32_t flags);
float MSG_ReadAngle16(uint32_t flags);  //johnfitz

//============================================================================

/* use our own copies of strlcpy and strlcat taken from OpenBSD */
size_t q_strlcpy(char *dst, const char *src, size_t size);
size_t q_strlcat(char *dst, const char *src, size_t size);

/* locale-insensitive strcasecmp replacement functions: */
int32_t q_strcasecmp(const char * s1, const char * s2);
int32_t q_strncasecmp(const char *s1, const char *s2, size_t n);

/* locale-insensitive case-insensitive alternative to strstr */
char *q_strcasestr(const char *haystack, const char *needle);

/* locale-insensitive strlwr/upr replacement functions: */
char *q_strlwr(char *str);
char *q_strupr(char *str);

/* snprintf, vsnprintf : always use our versions. */
int32_t q_snprintf(char *str, size_t size, const char *format, ...) FUNC_PRINTF(3, 4);
int32_t q_vsnprintf(char *str, size_t size, const char *format, va_list args) FUNC_PRINTF(3, 0);

//============================================================================

extern  char            com_token[1024];
extern  bool    com_eof;

const char *COM_Parse(const char *data);


extern  int32_t         com_argc;
extern  char    **com_argv;

extern  int32_t         safemode;
/* safe mode: in true, the engine will behave as if one
   of these arguments were actually on the command line:
   -nosound, -nocdaudio, -nomidi, -stdvid, -dibonly,
   -nomouse, -nojoy, -nolan
 */

int32_t COM_CheckParm(const char *parm);

void COM_InitArgv(int32_t argc, char **argv);
void COM_InitFilesystem(void);

const char *COM_SkipPath(const char *pathname);
void COM_StripExtension(const char *in, char *out, size_t outsize);
void COM_FileBase(const char *in, char *out, size_t outsize);
void COM_AddExtension(char *path, const char *extension, size_t len);
const char *COM_FileGetExtension(const char *in);  /* doesn't return NULL */
void COM_ExtractExtension(const char *in, char *out, size_t outsize);
void COM_CreatePath(char *path);

char *va(const char *format, ...) FUNC_PRINTF(1, 2);
// does a varargs printf into a temp buffer


//============================================================================

// QUAKEFS
typedef struct
{
	char    name[MAX_QPATH];
	int32_t         filepos, filelen;
} packfile_t;

typedef struct pack_s
{
	char    filename[MAX_OSPATH];
	int32_t         handle;
	int32_t         numfiles;
	packfile_t      *files;
} pack_t;

typedef struct searchpath_s
{
	uint32_t path_id;       // identifier assigned to the game directory
	// Note that <install_dir>/game1 and
	// <userdir>/game1 have the same id.
	char    filename[MAX_OSPATH];
	pack_t  *pack;                  // only one of filename / pack will be used
	struct searchpath_s     *next;
} searchpath_t;

extern searchpath_t *com_searchpaths;
extern searchpath_t *com_base_searchpaths;

extern int32_t com_filesize;
struct cache_user_s;

extern  char    com_basedir[MAX_OSPATH];
extern  char    com_gamedir[MAX_OSPATH];
extern  int32_t file_from_pak;  // global indicating that file came from a pak

void COM_WriteFile(const char *filename, const void *data, int32_t len);
int32_t COM_OpenFile(const char *filename, int32_t *handle, uint32_t *path_id);
int32_t COM_FOpenFile(const char *filename, FILE **file, uint32_t *path_id);
bool COM_FileExists(const char *filename, uint32_t *path_id);
void COM_CloseFile(int32_t h);

// these procedures open a file using COM_FindFile and loads it into a proper
// buffer. the buffer is allocated with a total size of com_filesize + 1. the
// procedures differ by their buffer allocation method.
byte *COM_LoadStackFile(const char *path, void *buffer, int32_t bufsize,
                        uint32_t *path_id);
// uses the specified stack stack buffer with the specified size
// of bufsize. if bufsize is too short, uses temp hunk. the bufsize
// must include the +1
byte *COM_LoadTempFile(const char *path, uint32_t *path_id);
// allocates the buffer on the temp hunk.
byte *COM_LoadHunkFile(const char *path, uint32_t *path_id);
// allocates the buffer on the hunk.
byte *COM_LoadZoneFile(const char *path, uint32_t *path_id);
// allocates the buffer on the zone.
byte *COM_LoadCacheFile(const char *path, struct cache_user_s *cu,
                        uint32_t *path_id);
// uses cache mem for allocating the buffer.
byte *COM_LoadMallocFile(const char *path, uint32_t *path_id);
// allocates the buffer on the system mem (malloc).

// Opens the given path directly, ignoring search paths.
// Returns NULL on failure, or else a '\0'-terminated malloc'ed buffer.
// Loads in "t" mode so CRLF to LF translation is performed on Windows.
byte *COM_LoadMallocFile_TextMode_OSPath(const char *path, long *len_out);

// Attempts to parse an int32_t, followed by a newline.
// Returns advanced buffer position.
// Doesn't signal parsing failure, but this is not needed for savegame loading.
const char *COM_ParseIntNewline(const char *buffer, int32_t *value);

// Attempts to parse a float followed by a newline.
// Returns advanced buffer position.
const char *COM_ParseFloatNewline(const char *buffer, float *value);

// Parse a string of non-whitespace into com_token, then tries to consume a
// newline. Returns advanced buffer position.
const char *COM_ParseStringNewline(const char *buffer);

/* The following FS_*() stdio replacements are necessary if one is
 * to perform non-sequential reads on files reopened on pak files
 * because we need the bookkeeping about file start/end positions.
 * Allocating and filling in the fshandle_t structure is the users'
 * responsibility when the file is initially opened. */

typedef struct
{
	FILE *file;
	bool pak;       /* is the file read from a pak */
	long start;     /* file or data start position */
	long length;    /* file or data size */
	long pos;       /* current position relative to start */
} fshandle_t;

size_t FS_fread(void *ptr, size_t size, size_t nmemb, fshandle_t *fh);
int32_t FS_fseek(fshandle_t *fh, long offset, int32_t whence);
long FS_ftell(fshandle_t *fh);
void FS_rewind(fshandle_t *fh);
int32_t FS_feof(fshandle_t *fh);
int32_t FS_ferror(fshandle_t *fh);
int32_t FS_fclose(fshandle_t *fh);
int32_t FS_fgetc(fshandle_t *fh);
char *FS_fgets(char *s, int32_t size, fshandle_t *fh);
long FS_filelength(fshandle_t *fh);

extern bool standard_quake, rogue, hipnotic;

struct cvar_s;

#endif
