/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2010-2014 QuakeSpasm developers

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "q_defs.h"

typedef struct
{
	int32_t         s;
	dfunction_t     *f;
} prstack_t;

#define MAX_STACK_DEPTH         64      /* was 32 */
static prstack_t        pr_stack[MAX_STACK_DEPTH];
static int32_t          pr_depth;

#define LOCALSTACK_SIZE         2048
static int32_t          localstack[LOCALSTACK_SIZE];
static int32_t          localstack_used;

bool    pr_trace;
dfunction_t     *pr_xfunction;
int32_t         pr_xstatement;
int32_t         pr_argc;

static const char *pr_opnames[] =
{
#define pr_ops__xmac(x) #x,
#include "pr_ops.h"
};

const char *PR_GlobalString(int32_t ofs);
const char *PR_GlobalStringNoContents(int32_t ofs);


//=============================================================================

/*
=================
PR_PrintStatement
=================
*/
static void PR_PrintStatement(dstatement_t *s)
{
	size_t i;

	if((uint32_t)s->op < arraysizeof(pr_opnames))
	{
		Con_Printf("%s ", pr_opnames[s->op]);
		i = strlen(pr_opnames[s->op]);
		for(; i < 10; i++)
			Con_Printf(" ");
	}

	if(s->op == OP_IF || s->op == OP_IFNOT)
		Con_Printf("%sbranch %" PRIi32, PR_GlobalString(s->a), s->b);
	else if(s->op == OP_GOTO)
	{
		Con_Printf("branch %" PRIi32, s->a);
	}
	else if((uint32_t)(s->op - OP_STORE_F) < 6)
	{
		Con_Printf("%s", PR_GlobalString(s->a));
		Con_Printf("%s", PR_GlobalStringNoContents(s->b));
	}
	else
	{
		if(s->a)
			Con_Printf("%s", PR_GlobalString(s->a));
		if(s->b)
			Con_Printf("%s", PR_GlobalString(s->b));
		if(s->c)
			Con_Printf("%s", PR_GlobalStringNoContents(s->c));
	}
	Con_Printf("\n");
}

/*
============
PR_StackTrace
============
*/
static void PR_StackTrace(void)
{
	int32_t         i;
	dfunction_t     *f;

	if(pr_depth == 0)
	{
		Con_Printf("<NO STACK>\n");
		return;
	}

	pr_stack[pr_depth].f = pr_xfunction;
	for(i = pr_depth; i >= 0; i--)
	{
		f = pr_stack[i].f;
		if(!f)
		{
			Con_Printf("<NO FUNCTION>\n");
		}
		else
		{
			Con_Printf("%12s : %s\n", PR_GetString(f->s_file), PR_GetString(f->s_name));
		}
	}
}


/*
============
PR_Profile_f

============
*/
void PR_Profile_f(void)
{
	int32_t         i, num;
	int32_t         pmax;
	dfunction_t     *f, *best;

	if(!sv.active)
		return;

	num = 0;
	do
	{
		pmax = 0;
		best = NULL;
		for(i = 0; i < progs.numfunctions; i++)
		{
			f = &pr_functions[i];
			if(f->profile > pmax)
			{
				pmax = f->profile;
				best = f;
			}
		}
		if(best)
		{
			if(num < 10)
				Con_Printf("%7" PRIi32 " %s\n", best->profile, PR_GetString(best->s_name));
			num++;
			best->profile = 0;
		}
	}
	while(best);
}


/*
============
PR_RunError

Aborts the currently executing function
============
*/
void PR_RunError(const char *error, ...)
{
	va_list argptr;
	char    string[1024];

	va_start(argptr, error);
	q_vsnprintf(string, sizeof(string), error, argptr);
	va_end(argptr);

	PR_PrintStatement(pr_statements + pr_xstatement);
	PR_StackTrace();

	Con_Printf("%s\n", string);

	pr_depth = 0;   // dump the stack so host_error can shutdown functions

	Host_Error("Program error");
}

/*
====================
PR_EnterFunction

Returns the new program statement counter
====================
*/
static int32_t PR_EnterFunction(dfunction_t *f)
{
	int32_t i, j, c, o;

	pr_stack[pr_depth].s = pr_xstatement;
	pr_stack[pr_depth].f = pr_xfunction;
	pr_depth++;
	if(pr_depth >= MAX_STACK_DEPTH)
		PR_RunError("stack overflow");

	// save off any locals that the new function steps on
	c = f->locals;
	if(localstack_used + c > LOCALSTACK_SIZE)
		PR_RunError("PR_ExecuteProgram: locals stack overflow\n");

	for(i = 0; i < c ; i++)
		localstack[localstack_used + i] = G_Int(f->parm_start + i);
	localstack_used += c;

	// copy parameters
	o = f->parm_start;
	for(i = 0; i < f->numparms; i++)
	{
		for(j = 0; j < f->parm_size[i]; j++)
		{
			G_Int(o) = G_Int(GBL_PARM0 + i * 3 + j);
			o++;
		}
	}

	pr_xfunction = f;
	return f->first_statement - 1;  // offset the s++
}

/*
====================
PR_LeaveFunction
====================
*/
static int32_t PR_LeaveFunction(void)
{
	int32_t i, c;

	if(pr_depth <= 0)
		Host_Error("prog stack underflow");

	// Restore locals from the stack
	c = pr_xfunction->locals;
	localstack_used -= c;
	if(localstack_used < 0)
		PR_RunError("PR_ExecuteProgram: locals stack underflow");

	for(i = 0; i < c; i++)
		G_Int(pr_xfunction->parm_start + i) = localstack[localstack_used + i];

	// up stack
	pr_depth--;
	pr_xfunction = pr_stack[pr_depth].f;
	return pr_stack[pr_depth].s;
}


/*
====================
PR_ExecuteProgram

The interpretation main loop
====================
*/
void PR_ExecuteProgram(func_t fnum)
{
#if HAVE_COMPUTED_GOTO
#define CASE(op)    oper_##op:
#define BREAK       goto oper_end;
#define GOTO_OP(op) goto *oper_list[op];
#define END_GOTO_OP oper_end:;
#else
#define CASE(op)    case op:
#define BREAK       break;
#define GOTO_OP(op) switch(op) {
#define END_GOTO_OP }
#endif

#define OPA G_Eval((uint16_t)st->a)
#define OPB G_Eval((uint16_t)st->b)
#define OPC G_Eval((uint16_t)st->c)
#define PTR ((eval_t *)ProgEdict(OPB->edict))
#define EDI ProgEdict(OPA->edict)

#define MakeOpArithF(op, exp) \
		CASE(op) \
			OPC->flt = OPA->flt exp OPB->flt; \
			BREAK

#define MakeOpArithV(op, exp) \
		CASE(op) \
			OPC->vec[0] = OPA->vec[0] exp OPB->vec[0]; \
			OPC->vec[1] = OPA->vec[1] exp OPB->vec[1]; \
			OPC->vec[2] = OPA->vec[2] exp OPB->vec[2]; \
			BREAK

#define MakeOpArithB(op, exp) \
		CASE(op) \
			OPC->flt = (int32_t)OPA->flt exp (int32_t)OPB->flt; \
			BREAK

#if HAVE_COMPUTED_GOTO
	static void const *const oper_list[] = {
#define pr_ops__xmac(x) &&oper_OP_##x,
#include "pr_ops.h"
	};
#endif

	dstatement_t *st;
	dfunction_t  *f, *newf;
	int32_t      profile, startprofile;
	edict_t      *ed;
	int32_t      exitdepth;

	if(!fnum || fnum >= progs.numfunctions)
	{
		if(G_PEdict(GBL_self))
			ED_Print(ProgEdict(G_PEdict(GBL_self)));
		Host_Error("PR_ExecuteProgram: NULL function");
	}

	f = &pr_functions[fnum];

	pr_trace = false;

	// make a stack frame
	exitdepth = pr_depth;

	st = &pr_statements[PR_EnterFunction(f)];
	startprofile = profile = 0;

	for(;;)
	{
		/* next statement */
		st++;

		if(++profile > 100000)
		{
			pr_xstatement = st - pr_statements;
			PR_RunError("runaway loop error");
		}

		if(pr_trace)
			PR_PrintStatement(st);

		GOTO_OP(st->op)
		MakeOpArithF(OP_ADD_F, +)
		MakeOpArithF(OP_SUB_F, -)
		MakeOpArithF(OP_MUL_F, *)
		MakeOpArithF(OP_DIV_F, /)

		MakeOpArithV(OP_ADD_V, +)
		MakeOpArithV(OP_SUB_V, -)

		MakeOpArithB(OP_BITAND, &)
		MakeOpArithB(OP_BITOR,  |)

		CASE(OP_MUL_V)  OPC->flt = DotProduct(OPA->vec, OPB->vec); BREAK
		CASE(OP_MUL_FV) VectorScale(OPB->vec, OPA->flt, OPC->vec); BREAK
		CASE(OP_MUL_VF) VectorScale(OPA->vec, OPB->flt, OPC->vec); BREAK

		MakeOpArithF(OP_GE,   >=)
		MakeOpArithF(OP_LE,   <=)
		MakeOpArithF(OP_GT,   >)
		MakeOpArithF(OP_LT,   <)
		MakeOpArithF(OP_AND,  &&)
		MakeOpArithF(OP_OR,   ||)
		MakeOpArithF(OP_EQ_F, ==)
		MakeOpArithF(OP_NE_F, !=)

		CASE(OP_NOT_F)
			OPC->flt = !OPA->flt;
			BREAK
		CASE(OP_NOT_V)
			OPC->flt = !OPA->vec[0] && !OPA->vec[1] && !OPA->vec[2];
			BREAK
		CASE(OP_NOT_S)
			OPC->flt = !OPA->string || !*PR_GetString(OPA->string);
			BREAK
		CASE(OP_NOT_FNC)
			OPC->flt = !OPA->func;
			BREAK
		CASE(OP_NOT_ENT)
			OPC->flt = EDI == sv.edicts;
			BREAK

		CASE(OP_EQ_V)
			OPC->flt = OPA->vec[0] == OPB->vec[0] &&
			           OPA->vec[1] == OPB->vec[1] &&
			           OPA->vec[2] == OPB->vec[2];
			BREAK
		CASE(OP_EQ_S)
			OPC->flt = !strcmp(PR_GetString(OPA->string), PR_GetString(OPB->string));
			BREAK
		CASE(OP_EQ_E)   OPC->flt = OPA->edict == OPB->edict; BREAK
		CASE(OP_EQ_FNC) OPC->flt = OPA->func  == OPB->func;  BREAK

		CASE(OP_NE_V)
			OPC->flt = OPA->vec[0] != OPB->vec[0] ||
			           OPA->vec[1] != OPB->vec[1] ||
			           OPA->vec[2] != OPB->vec[2];
			BREAK
		CASE(OP_NE_S)
			OPC->flt = strcmp(PR_GetString(OPA->string), PR_GetString(OPB->string));
			BREAK
		CASE(OP_NE_E)   OPC->flt = OPA->edict != OPB->edict; BREAK
		CASE(OP_NE_FNC) OPC->flt = OPA->func  != OPB->func;  BREAK

		CASE(OP_STORE_F)   OPB->flt    = OPA->flt;         BREAK
		CASE(OP_STORE_ENT) OPB->edict  = OPA->edict;       BREAK
		CASE(OP_STORE_FLD) OPB->field  = OPA->field;       BREAK
		CASE(OP_STORE_S)   OPB->string = OPA->string;      BREAK
		CASE(OP_STORE_FNC) OPB->func   = OPA->func;        BREAK
		CASE(OP_STORE_V)   VectorCopy(OPA->vec, OPB->vec); BREAK

		CASE(OP_STOREP_F)   PTR->flt    = OPA->flt;         BREAK
		CASE(OP_STOREP_ENT) PTR->edict  = OPA->edict;       BREAK
		CASE(OP_STOREP_FLD) PTR->field  = OPA->field;       BREAK
		CASE(OP_STOREP_S)   PTR->string = OPA->string;      BREAK
		CASE(OP_STOREP_FNC) PTR->func   = OPA->func;        BREAK
		CASE(OP_STOREP_V)   VectorCopy(OPA->vec, PTR->vec); BREAK

		CASE(OP_ADDRESS)
			ed = EDI;
#if defined(PARANOID)
			NumForEdict(ed);      // Make sure it's in range
#endif
			if(ed == (edict_t *)sv.edicts && sv.state == ss_active)
			{
				pr_xstatement = st - pr_statements;
				PR_RunError("assignment to world entity");
			}
			OPC->field = EdictProg(&ED_PField(ed, OPB->field));
			BREAK

		CASE(OP_LOAD_F)   OPC->flt    = ED_Float(EDI, OPB->field);          BREAK
		CASE(OP_LOAD_FLD) OPC->field  = ED_PField(EDI, OPB->field);         BREAK
		CASE(OP_LOAD_ENT) OPC->edict  = ED_PEdict(EDI, OPB->field);         BREAK
		CASE(OP_LOAD_S)   OPC->string = ED_RString(EDI, OPB->field);        BREAK
		CASE(OP_LOAD_FNC) OPC->func   = ED_Func(EDI, OPB->field);           BREAK
		CASE(OP_LOAD_V)   VectorCopy(ED_Vector(EDI, OPB->field), OPC->vec); BREAK

		CASE(OP_IFNOT) if(!OPA->_int) st += st->b - 1; BREAK
		CASE(OP_IF)    if(OPA->_int)  st += st->b - 1; BREAK
		CASE(OP_GOTO)                 st += st->a - 1; BREAK

		CASE(OP_CALL0)
		CASE(OP_CALL1)
		CASE(OP_CALL2)
		CASE(OP_CALL3)
		CASE(OP_CALL4)
		CASE(OP_CALL5)
		CASE(OP_CALL6)
		CASE(OP_CALL7)
		CASE(OP_CALL8)
			pr_xfunction->profile += profile - startprofile;
			startprofile = profile;
			pr_xstatement = st - pr_statements;
			pr_argc = st->op - OP_CALL0;
			if(!OPA->func)
				PR_RunError("NULL function");
			newf = &pr_functions[OPA->func];
			if(newf->first_statement < 0)
			{
				// Built-in function
				int32_t i = -newf->first_statement;
				if(i >= pr_numbuiltins)
					PR_RunError("Bad builtin call number %" PRIi32, i);
				pr_builtins[i]();
				BREAK
			}
			// Normal function
			st = &pr_statements[PR_EnterFunction(newf)];
			BREAK

		CASE(OP_DONE)
		CASE(OP_RETURN)
			pr_xfunction->profile += profile - startprofile;
			startprofile = profile;
			pr_xstatement = st - pr_statements;
			G_Float(GBL_RETURN + 0) = G_Float((uint16_t)st->a + 0);
			G_Float(GBL_RETURN + 1) = G_Float((uint16_t)st->a + 1);
			G_Float(GBL_RETURN + 2) = G_Float((uint16_t)st->a + 2);
			st = &pr_statements[PR_LeaveFunction()];
			if(pr_depth == exitdepth)
				return;
			BREAK

		CASE(OP_STATE)
			ed = ProgEdict(G_PEdict(GBL_self));
			ED_Float(ed, ED_nextthink) = G_Float(GBL_time) + 0.1;
			ED_Float(ed, ED_frame) = OPA->flt;
			ED_Func(ed, ED_think) = OPB->func;
			BREAK
		END_GOTO_OP
	}
}

